@@ -7,6 +7,9 @@
 // Various tools for parsing LLVM. Utilities of various sorts, that are
 // specific to Emscripten (and hence not in utility.js).
 
+const FOUR_GB = 4 * 1024 * 1024 * 1024;
+const FLOAT_TYPES = set('float', 'double');
+
 let currentlyParsedFilename = '';
 
 // Does simple 'macro' substitution, using Django-like syntax,
@@ -31,43 +34,69 @@ function processMacros(text) {
 // Param filenameHint can be passed as a description to identify the file that is being processed, used
 // to locate errors for reporting and for html files to stop expansion between <style> and </style>.
 function preprocess(text, filenameHint) {
+  if (EXPORT_ES6 && USE_ES6_IMPORT_META) {
+    // `eval`, Terser and Closure don't support module syntax; to allow it,
+    // we need to temporarily replace `import.meta` usages with placeholders
+    // during preprocess phase, and back after all the other ops.
+    // See also: `phase_final_emitting` in emcc.py.
+    text = text.replace(/\bimport\.meta\b/g, 'EMSCRIPTEN$IMPORT$META');
+  }
+
+  const IGNORE = 0;
+  const SHOW = 1;
+  // This state is entered after we have shown one of the block of an if/elif/else sequence.
+  // Once we enter this state we dont show any blocks or evaluate any
+  // conditions until the sequence ends.
+  const IGNORE_ALL = 2;
+  const showStack = [];
+  const showCurrentLine = () => showStack.every((x) => x == SHOW);
+
   currentlyParsedFilename = filenameHint;
+  const fileExt = (filenameHint) ? filenameHint.split('.').pop().toLowerCase() : '';
+  const isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
+  let inStyle = false;
+  const lines = text.split('\n');
+  let ret = '';
+  let emptyLine = false;
+
   try {
-    const fileExt = (filenameHint) ? filenameHint.split('.').pop().toLowerCase() : '';
-    const isHtml = (fileExt === 'html' || fileExt === 'htm') ? true : false;
-    let inStyle = false;
-    const lines = text.split('\n');
-    let ret = '';
-    const showStack = [];
-    let emptyLine = false;
     for (let i = 0; i < lines.length; i++) {
       let line = lines[i];
       try {
         if (line[line.length - 1] === '\r') {
           line = line.substr(0, line.length - 1); // Windows will have '\r' left over from splitting over '\r\n'
         }
-        if (isHtml && line.indexOf('<style') !== -1 && !inStyle) {
+        if (isHtml && line.includes('<style') && !inStyle) {
           inStyle = true;
         }
-        if (isHtml && line.indexOf('</style') !== -1 && inStyle) {
+        if (isHtml && line.includes('</style') && inStyle) {
           inStyle = false;
         }
 
         if (!inStyle) {
           const trimmed = line.trim();
-          if (trimmed[0] === '#') {
+          if (trimmed.startsWith('#')) {
             const first = trimmed.split(' ', 1)[0];
-            if (first == '#if' || first == '#ifdef') {
+            if (first == '#if' || first == '#ifdef' || first == '#elif') {
               if (first == '#ifdef') {
                 warn('warning: use of #ifdef in js library.  Use #if instead.');
               }
+              if (first == '#elif') {
+                const curr = showStack.pop();
+                if (curr == SHOW || curr == IGNORE_ALL) {
+                  // If we showed to previous block we enter the IGNORE_ALL state
+                  // and stay there until endif is seen
+                  showStack.push(IGNORE_ALL);
+                  continue;
+                }
+              }
               const after = trimmed.substring(trimmed.indexOf(' '));
               const truthy = !!eval(after);
-              showStack.push(truthy);
+              showStack.push(truthy ? SHOW : IGNORE);
             } else if (first === '#include') {
-              if (showStack.indexOf(false) === -1) {
+              if (showCurrentLine()) {
                 let filename = line.substr(line.indexOf(' ') + 1);
-                if (filename.indexOf('"') === 0) {
+                if (filename.startsWith('"')) {
                   filename = filename.substr(1, filename.length - 2);
                 }
                 const included = read(filename);
@@ -80,7 +109,12 @@ function preprocess(text, filenameHint) {
               }
             } else if (first === '#else') {
               assert(showStack.length > 0);
-              showStack.push(!showStack.pop());
+              const curr = showStack.pop();
+              if (curr == IGNORE) {
+                showStack.push(SHOW);
+              } else {
+                showStack.push(IGNORE);
+              }
             } else if (first === '#endif') {
               assert(showStack.length > 0);
               showStack.pop();
@@ -88,7 +122,7 @@ function preprocess(text, filenameHint) {
               throw new Error(`Unknown preprocessor directive on line ${i}: ``${line}```);
             }
           } else {
-            if (showStack.indexOf(false) === -1) {
+            if (showCurrentLine()) {
               // Never emit more than one empty line at a time.
               if (emptyLine && !line) {
                 continue;
@@ -102,7 +136,7 @@ function preprocess(text, filenameHint) {
             }
           }
         } else { // !inStyle
-          if (showStack.indexOf(false) === -1) {
+          if (showCurrentLine()) {
             ret += line + '\n';
           }
         }
@@ -119,27 +153,6 @@ no matching #endif found (${showStack.length$}' unmatched preprocessing directiv
   }
 }
 
-function removePointing(type, num) {
-  if (num === 0) return type;
-  assert(type.substr(type.length - (num ? num : 1)).replace(/\*/g, '') === '');
-  // , 'Error in removePointing with ' + [type, num, type.substr(type.length-(num ? num : 1))]);
-  return type.substr(0, type.length - (num ? num : 1));
-}
-
-function pointingLevels(type) {
-  if (!type) return 0;
-  let ret = 0;
-  const len1 = type.length - 1;
-  while (type[len1 - ret] && type[len1 - ret] === '*') {
-    ret++;
-  }
-  return ret;
-}
-
-function removeAllPointing(type) {
-  return removePointing(type, pointingLevels(type));
-}
-
 // Returns true if ident is a niceIdent (see toNiceIdent). Also allow () and spaces.
 function isNiceIdent(ident, loose) {
   return /^\(?[$_]+[\w$_\d ]*\)?$/.test(ident);
@@ -152,48 +165,18 @@ function needsQuoting(ident) {
   return true;
 }
 
-function isStructPointerType(type) {
-  // This test is necessary for clang - in llvm-gcc, we
-  // could check for %struct. The downside is that %1 can
-  // be either a variable or a structure, and we guess it is
-  // a struct, which can lead to |call i32 %5()| having
-  // |%5()| as a function call (like |i32 (i8*)| etc.). So
-  // we must check later on, in call(), where we have more
-  // context, to differentiate such cases.
-  // A similar thing happens in isStructType()
-  return !Compiletime.isNumberType(type) && type[0] == '%';
-}
+const POINTER_SIZE = MEMORY64 ? 8 : 4;
+const POINTER_BITS = POINTER_SIZE * 8;
+const POINTER_TYPE = 'i' + POINTER_BITS;
+
+const SIZE_TYPE = POINTER_TYPE;
 
 function isPointerType(type) {
   return type[type.length - 1] == '*';
 }
 
-function isArrayType(type) {
-  return /^\[\d+\ x\ (.*)\]/.test(type);
-}
-
-function isStructType(type) {
-  if (isPointerType(type)) return false;
-  if (isArrayType(type)) return true;
-  if (/<?\{ ?[^}]* ?\}>?/.test(type)) return true; // { i32, i8 } etc. - anonymous struct types
-  // See comment in isStructPointerType()
-  return type[0] == '%';
-}
-
-function isVectorType(type) {
-  return type[type.length - 1] === '>';
-}
-
-function isStructuralType(type) {
-  return /^\{ ?[^}]* ?\}$/.test(type); // { i32, i8 } etc. - anonymous struct types
-}
-
-function getStructuralTypeParts(type) { // split { i32, i8 } etc. into parts
-  return type.replace(/[ {}]/g, '').split(',');
-}
-
-function getStructuralTypePartBits(part) {
-  return Math.ceil((getBits(part) || 32) / 32) * 32; // simple 32-bit alignment. || 32 is for pointers
+function pointerT(x) {
+  return MEMORY64 ? `BigInt(${x})` : x;
 }
 
 function isIntImplemented(type) {
@@ -202,170 +185,16 @@ function isIntImplemented(type) {
 
 // Note: works for iX types and structure types, not pointers (even though they are implemented as ints)
 function getBits(type, allowPointers) {
-  if (allowPointers && isPointerType(type)) return 32;
+  if (allowPointers && isPointerType(type)) return POINTER_SIZE;
   if (!type) return 0;
   if (type[0] == 'i') {
     const left = type.substr(1);
     if (!isNumber(left)) return 0;
     return parseInt(left);
   }
-  if (isStructuralType(type)) {
-    return sum(getStructuralTypeParts(type).map(getStructuralTypePartBits));
-  }
-  if (isStructType(type)) {
-    const typeData = Types.types[type];
-    if (typeData === undefined) return 0;
-    return typeData.flatSize * 8;
-  }
   return 0;
 }
 
-function isVoidType(type) {
-  return type == 'void';
-}
-
-// Detects a function definition, ([...|type,[type,...]])
-function isFunctionDef(token, out) {
-  const text = token.text;
-  const nonPointing = removeAllPointing(text);
-  if (nonPointing[0] != '(' || nonPointing.substr(-1) != ')') {
-    return false;
-  }
-  if (nonPointing === '()') return true;
-  if (!token.tokens) return false;
-  let fail = false;
-  const segments = splitTokenList(token.tokens);
-  segments.forEach((segment) => {
-    const subtext = segment[0].text;
-    fail = fail || segment.length > 1 || !(isType(subtext) || subtext == '...');
-  });
-  if (out) {
-    out.segments = segments;
-    out.numArgs = segments.length;
-  }
-  return !fail;
-}
-
-function isPossiblyFunctionType(type) {
-  // A quick but unreliable way to see if something is a function type. Yes is just 'maybe', no is definite.
-  const len = type.length;
-  return type[len - 2] == ')' && type[len - 1] == '*';
-}
-
-function isFunctionType(type, out) {
-  if (!isPossiblyFunctionType(type)) return false;
-  type = type.substr(0, type.length - 1); // remove final '*'
-  const firstOpen = type.indexOf('(');
-  if (firstOpen <= 0) return false;
-  type = type.replace(/"[^"]+"/g, '".."');
-  const lastOpen = type.lastIndexOf('(');
-  let returnType;
-  if (firstOpen == lastOpen) {
-    returnType = getReturnType(type);
-    if (!isType(returnType)) return false;
-  } else {
-    returnType = 'i8*'; // some pointer type, no point in analyzing further
-  }
-  if (out) out.returnType = returnType;
-  // find ( that starts the arguments
-  let depth = 0;
-  let i = type.length - 1;
-  let argText = null;
-  while (i >= 0) {
-    const curr = type[i];
-    if (curr == ')') depth++;
-    else if (curr == '(') {
-      depth--;
-      if (depth == 0) {
-        argText = type.substr(i);
-        break;
-      }
-    }
-    i--;
-  }
-  assert(argText);
-  return isFunctionDef({text: argText, tokens: tokenize(argText.substr(1, argText.length - 2))}, out);
-}
-
-function getReturnType(type) {
-  // the type of a call can be either the return value, or the entire function. ** or more means it is a return value
-  if (pointingLevels(type) > 1) return '*';
-  const lastOpen = type.lastIndexOf('(');
-  if (lastOpen > 0) {
-    // handle things like   void (i32)* (i32, void (i32)*)*
-    const closeStar = type.indexOf(')*');
-    if (closeStar > 0 && closeStar < type.length - 2) lastOpen = closeStar + 3;
-    return type.substr(0, lastOpen - 1);
-  }
-  return type;
-}
-
-const isTypeCache = {}; // quite hot, optimize as much as possible
-
-function isType(type) {
-  if (type in isTypeCache) return isTypeCache[type];
-  const ret = isPointerType(type) || isVoidType(type) || Compiletime.isNumberType(type) || isStructType(type) || isFunctionType(type);
-  isTypeCache[type] = ret;
-  return ret;
-}
-
-const SPLIT_TOKEN_LIST_SPLITTERS = set(',', 'to'); // 'to' can separate parameters as well...
-
-// Splits a list of tokens separated by commas. For example, a list of arguments in a function call
-function splitTokenList(tokens) {
-  if (tokens.length == 0) return [];
-  if (!tokens.slice) tokens = tokens.tokens;
-  const ret = [];
-  const seg = [];
-  for (let i = 0; i < tokens.length; i++) {
-    const token = tokens[i];
-    if (token.text in SPLIT_TOKEN_LIST_SPLITTERS) {
-      ret.push(seg);
-      seg = [];
-    } else if (token.text == ';') {
-      ret.push(seg);
-      return ret;
-    } else {
-      seg.push(token);
-    }
-  }
-  if (seg.length) ret.push(seg);
-  return ret;
-}
-
-function _IntToHex(x) {
-  assert(x >= 0 && x <= 15);
-  if (x <= 9) {
-    return String.fromCharCode('0'.charCodeAt(0) + x);
-  }
-  return String.fromCharCode('A'.charCodeAt(0) + x - 10);
-}
-
-function IEEEUnHex(stringy) {
-  stringy = stringy.substr(2); // leading '0x';
-  if (stringy.replace(/0/g, '') === '') return 0;
-  while (stringy.length < 16) stringy = '0' + stringy;
-  assert(stringy.length === 16, 'Can only unhex 16-digit double numbers, nothing platform-specific'); // |long double| might cause this
-  const top = eval('0x' + stringy[0]);
-  const neg = !!(top & 8); // sign
-  if (neg) {
-    stringy = _IntToHex(top & ~8) + stringy.substr(1);
-  }
-  let a = eval('0x' + stringy.substr(0, 8)); // top half
-  const b = eval('0x' + stringy.substr(8)); // bottom half
-  let e = a >> ((52 - 32) & 0x7ff); // exponent
-  a = a & 0xfffff;
-  if (e === 0x7ff) {
-    if (a == 0 && b == 0) {
-      return neg ? '-Infinity' : 'Infinity';
-    }
-    return 'NaN';
-  }
-  e -= 1023; // offset
-  const absolute = ((((a | 0x100000) * 1.0) / Math.pow(2, 52 - 32)) * Math.pow(2, e)) + (((b * 1.0) / Math.pow(2, 52)) * Math.pow(2, e));
-  return (absolute * (neg ? -1 : 1)).toString();
-}
-
 // Given an expression like (VALUE=VALUE*2,VALUE<10?VALUE:t+1) , this will
 // replace VALUE with value. If value is not a simple identifier of a variable,
 // value will be replaced with tempVar.
@@ -447,7 +276,7 @@ function checkSafeHeap() {
 }
 
 function getHeapOffset(offset, type) {
-  if (Runtime.getNativeFieldSize(type) > 4 && type == 'i64') {
+  if (!WASM_BIGINT && Runtime.getNativeFieldSize(type) > 4 && type == 'i64') {
     // we emulate 64-bit integer values as 32 in asmjs-unknown-emscripten, but not double
     type = 'i32';
   }
@@ -462,7 +291,7 @@ function ensureDot(value) {
   // if already dotted, or Infinity or NaN, nothing to do here
   // if smaller than 1 and running js opts, we always need to force a coercion
   // (0.001 will turn into 1e-3, which has no .)
-  if ((value.indexOf('.') >= 0 || /[IN]/.test(value))) return value;
+  if ((value.includes('.') || /[IN]/.test(value))) return value;
   const e = value.indexOf('e');
   if (e < 0) return value + '.0';
   return value.substr(0, e) + '.0' + value.substr(e);
@@ -478,7 +307,7 @@ function asmEnsureFloat(value, type) {
     value = ensureDot(value);
     return 'Math.fround(' + value + ')';
   }
-  if (type in Compiletime.FLOAT_TYPES) {
+  if (type in FLOAT_TYPES) {
     return ensureDot(value);
   }
   return value;
@@ -487,7 +316,7 @@ function asmEnsureFloat(value, type) {
 function asmCoercion(value, type, signedness) {
   if (type == 'void') {
     return value;
-  } else if (type in Compiletime.FLOAT_TYPES) {
+  } else if (type in FLOAT_TYPES) {
     if (isNumber(value)) {
       return asmEnsureFloat(value, type);
     } else {
@@ -539,14 +368,6 @@ function makeSetTempDouble(i, type, value) {
 // See makeSetValue
 function makeGetValue(ptr, pos, type, noNeedFirst, unsigned, ignore, align, noSafe, forceAsm) {
   assert(!forceAsm, 'forceAsm is no longer supported');
-  if (isStructType(type)) {
-    const typeData = Types.types[type];
-    const ret = [];
-    for (let i = 0; i < typeData.fields.length; i++) {
-      ret.push('f' + i + ': ' + makeGetValue(ptr, pos + typeData.flatIndexes[i], typeData.fields[i], noNeedFirst, unsigned, 0, 0, noSafe));
-    }
-    return '{ ' + ret.join(', ') + ' }';
-  }
 
   if (type == 'double' && (align < 8)) {
     const setdouble1 = makeSetTempDouble(0, 'i32', makeGetValue(ptr, pos, 'i32', noNeedFirst, unsigned, ignore, align, noSafe));
@@ -587,10 +408,19 @@ function makeGetValue(ptr, pos, type, noNeedFirst, unsigned, ignore, align, noSa
   const offset = calcFastOffset(ptr, pos, noNeedFirst);
   if (SAFE_HEAP && !noSafe) {
     if (!ignore) {
-      return asmCoercion('SAFE_HEAP_LOAD' + ((type in Compiletime.FLOAT_TYPES) ? '_D' : '') + '(' + asmCoercion(offset, 'i32') + ', ' + Runtime.getNativeTypeSize(type) + ', ' + (!!unsigned + 0) + ')', type, unsigned ? 'u' : undefined);
+      return asmCoercion('SAFE_HEAP_LOAD' + ((type in FLOAT_TYPES) ? '_D' : '') + '(' + asmCoercion(offset, 'i32') + ', ' + Runtime.getNativeTypeSize(type) + ', ' + (!!unsigned + 0) + ')', type, unsigned ? 'u' : undefined);
     }
   }
-  return getHeapForType(type, unsigned) + '[' + getHeapOffset(offset, type) + ']';
+
+  const slab = getHeapForType(type, unsigned);
+  let ret = slab + '[' + getHeapOffset(offset, type) + ']';
+  if (slab.substr(slab.length - 2) == '64') {
+    ret = `Number(${ret})`;
+  }
+  if (forceAsm) {
+    ret = asmCoercion(ret, type);
+  }
+  return ret;
 }
 
 /**
@@ -614,25 +444,12 @@ function makeGetValue(ptr, pos, type, noNeedFirst, unsigned, ignore, align, noSa
 function makeSetValue(ptr, pos, value, type, noNeedFirst, ignore, align, noSafe, sep, forcedAlign) {
   assert(!forcedAlign, 'forcedAlign is no longer supported');
   sep = sep || ';';
-  if (isStructType(type)) {
-    const typeData = Types.types[type];
-    const ret = [];
-    // We can receive either an object - an object literal that was in the .ll - or a string,
-    // which is the ident of an aggregate struct
-    if (typeof value === 'string') {
-      value = range(typeData.fields.length).map((i) => value + '.f' + i);
-    }
-    for (let i = 0; i < typeData.fields.length; i++) {
-      ret.push(makeSetValue(ptr, getFastValue(pos, '+', typeData.flatIndexes[i]), value[i], typeData.fields[i], noNeedFirst, 0, 0, noSafe));
-    }
-    return ret.join('; ');
-  }
 
   if (type == 'double' && (align < 8)) {
     return '(' + makeSetTempDouble(0, 'double', value) + ',' +
             makeSetValue(ptr, pos, makeGetTempDouble(0, 'i32'), 'i32', noNeedFirst, ignore, align, noSafe, ',') + ',' +
             makeSetValue(ptr, getFastValue(pos, '+', Runtime.getNativeTypeSize('i32')), makeGetTempDouble(1, 'i32'), 'i32', noNeedFirst, ignore, align, noSafe, ',') + ')';
-  } else if (type == 'i64') {
+  } else if (!WASM_BIGINT && type == 'i64') {
     return '(tempI64 = [' + splitI64(value) + '],' +
             makeSetValue(ptr, pos, 'tempI64[0]', 'i32', noNeedFirst, ignore, align, noSafe, ',') + ',' +
             makeSetValue(ptr, getFastValue(pos, '+', Runtime.getNativeTypeSize('i32')), 'tempI64[1]', 'i32', noNeedFirst, ignore, align, noSafe, ',') + ')';
@@ -669,10 +486,15 @@ function makeSetValue(ptr, pos, value, type, noNeedFirst, ignore, align, noSafe,
   const offset = calcFastOffset(ptr, pos, noNeedFirst);
   if (SAFE_HEAP && !noSafe) {
     if (!ignore) {
-      return 'SAFE_HEAP_STORE' + ((type in Compiletime.FLOAT_TYPES) ? '_D' : '') + '(' + asmCoercion(offset, 'i32') + ', ' + asmCoercion(value, type) + ', ' + Runtime.getNativeTypeSize(type) + ')';
+      return 'SAFE_HEAP_STORE' + ((type in FLOAT_TYPES) ? '_D' : '') + '(' + asmCoercion(offset, 'i32') + ', ' + asmCoercion(value, type) + ', ' + Runtime.getNativeTypeSize(type) + ')';
     }
   }
-  return getHeapForType(type) + '[' + getHeapOffset(offset, type) + '] = ' + value;
+
+  const slab = getHeapForType(type);
+  if (slab == 'HEAPU64' || slab == 'HEAP64') {
+    value = `BigInt(${value})`;
+  }
+  return slab + '[' + getHeapOffset(offset, type) + '] = ' + value;
 }
 
 const UNROLL_LOOP_MAX = 8;
@@ -782,7 +604,7 @@ function getFastValue(a, op, b, type) {
   if (op === '*') {
     // We can't eliminate where a or b are 0 as that would break things for creating
     // a negative 0.
-    if ((aNumber === 0 || bNumber === 0) && !(type in Compiletime.FLOAT_TYPES)) {
+    if ((aNumber === 0 || bNumber === 0) && !(type in FLOAT_TYPES)) {
       return '0';
     } else if (aNumber === 1) {
       return b;
@@ -794,7 +616,7 @@ function getFastValue(a, op, b, type) {
         return `(${a}<<${shifts})`;
       }
     }
-    if (!(type in Compiletime.FLOAT_TYPES)) {
+    if (!(type in FLOAT_TYPES)) {
       // if guaranteed small enough to not overflow into a double, do a normal multiply
       // default is 32-bit multiply for things like getelementptr indexes
       const bits = getBits(type) || 32;
@@ -808,7 +630,7 @@ function getFastValue(a, op, b, type) {
     }
   } else if (op === '/') {
     // careful on floats, since 0*NaN is not 0
-    if (a === '0' && !(type in Compiletime.FLOAT_TYPES)) {
+    if (a === '0' && !(type in FLOAT_TYPES)) {
       return '0';
     } else if (b === 1) {
       return a;
@@ -829,13 +651,15 @@ function getFastValue(a, op, b, type) {
 
 function calcFastOffset(ptr, pos, noNeedFirst) {
   assert(!noNeedFirst);
+  if (typeof ptr == 'bigint') ptr = Number(ptr);
+  if (typeof pos == 'bigint') pos = Number(pos);
   return getFastValue(ptr, '+', pos, 'i32');
 }
 
 function getHeapForType(type, unsigned) {
   assert(type);
   if (isPointerType(type)) {
-    type = 'i32'; // Hardcoded 32-bit
+    type = POINTER_TYPE;
   }
   switch (type) {
     case 'i1':
@@ -843,9 +667,13 @@ function getHeapForType(type, unsigned) {
       return unsigned ? 'HEAPU8' : 'HEAP8';
     case 'i16':
       return unsigned ? 'HEAPU16' : 'HEAP16';
+    case 'i64':
+      if (WASM_BIGINT) {
+        return unsigned ? 'HEAPU64' : 'HEAP64';
+      }
+      // fall through
     case '<4 x i32>':
     case 'i32':
-    case 'i64':
       return unsigned ? 'HEAPU32' : 'HEAP32';
     case 'double':
       return 'HEAPF64';
@@ -856,25 +684,17 @@ function getHeapForType(type, unsigned) {
   assert(false, 'bad heap type: ' + type);
 }
 
-function makeGetTempRet0() {
-  return '(getTempRet0() | 0)';
-}
-
-function makeSetTempRet0(value) {
-  return 'setTempRet0((' + value + ') | 0)';
-}
-
 // Takes a pair of return values, stashes one in tempRet0 and returns the other.
 // Should probably be renamed to `makeReturn64` but keeping this old name in
 // case external JS library code uses this name.
 function makeStructuralReturn(values) {
   assert(values.length == 2);
-  return makeSetTempRet0(values[1]) + '; return ' + asmCoercion(values[0], 'i32');
+  return 'setTempRet0(' + values[1] + '); return ' + asmCoercion(values[0], 'i32');
 }
 
 function makeThrow(what) {
-  if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING == 1) {
-    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s DISABLE_EXCEPTION_CATCHING=0 or DISABLE_EXCEPTION_CATCHING=2 to catch."';
+  if (ASSERTIONS && DISABLE_EXCEPTION_CATCHING) {
+    what += ' + " - Exception catching is disabled, this exception cannot be caught. Compile with -s NO_DISABLE_EXCEPTION_CATCHING or -s EXCEPTION_CATCHING_ALLOWED=[..] to catch."';
     if (MAIN_MODULE) {
       what += ' + " (note: in dynamic linking, if a side module wants exceptions, the main module must be built with that support)"';
     }
@@ -883,10 +703,7 @@ function makeThrow(what) {
 }
 
 function makeSignOp(value, type, op, force, ignore) {
-  if (type == 'i64') {
-    return value; // these are always assumed to be two 32-bit unsigneds.
-  }
-  if (isPointerType(type)) type = 'i32'; // Pointers are treated as 32-bit ints
+  if (isPointerType(type)) type = POINTER_TYPE;
   if (!value) return value;
   let bits;
   let full;
@@ -992,40 +809,39 @@ function asmFFICoercion(value, type) {
 }
 
 function makeDynCall(sig, funcPtr) {
-  assert(sig.indexOf('j') == -1);
+  assert(!sig.includes('j'), 'Cannot specify 64-bit signatures ("j" in signature string) with makeDynCall!');
+
+  const returnExpr = (sig[0] == 'v') ? '' : 'return';
+
+  let args = [];
+  for (let i = 1; i < sig.length; ++i) {
+    args.push(`a${i}`);
+  }
+  args = args.join(', ');
+
   if (funcPtr === undefined) {
     printErr(`warning: ${currentlyParsedFilename}: \
 Legacy use of {{{ makeDynCall("${sig}") }}}(funcPtr, arg1, arg2, ...). \
 Starting from Emscripten 2.0.2 (Aug 31st 2020), syntax for makeDynCall has changed. \
 New syntax is {{{ makeDynCall("${sig}", "funcPtr") }}}(arg1, arg2, ...). \
 Please update to new syntax.`);
-    const ret = (sig[0] == 'v') ? 'return ' : '';
-    let args = [];
-    for (let i = 1; i < sig.length; ++i) {
-      args.push(`a${i}`);
-    }
-    args = args.join(', ');
 
     if (DYNCALLS) {
-      return `(function(cb, ${args}) { return getDynCaller("${sig}", cb)(${args}) })`;
+      return `(function(cb, ${args}) { ${returnExpr} getDynCaller("${sig}", cb)(${args}) })`;
     } else {
-      return `(function(cb, ${args}) { return wasmTable.get(cb)(${args}) })`;
+      return `(function(cb, ${args}) { ${returnExpr} getWasmTableEntry(cb)(${args}) })`;
     }
   }
+
   if (DYNCALLS) {
     const dyncall = exportedAsmFunc(`dynCall_${sig}`);
     if (sig.length > 1) {
-      let args = [];
-      for (let i = 1; i < sig.length; ++i) {
-        args.push(`a${i}`);
-      }
-      args = args.join(', ');
-      return `(function(${args}) { ${dyncall}.apply(null, [${funcPtr}, ${args}]); })`;
+      return `(function(${args}) { ${returnExpr} ${dyncall}.apply(null, [${funcPtr}, ${args}]); })`;
     } else {
-      return `(function() { ${dyncall}.call(null, ${funcPtr}); })`;
+      return `(function() { ${returnExpr} ${dyncall}.call(null, ${funcPtr}); })`;
     }
   } else {
-    return `wasmTable.get(${funcPtr})`;
+    return `getWasmTableEntry(${funcPtr})`;
   }
 }
 
@@ -1054,9 +870,12 @@ function addAtInit(code) {
   ATINITS.push(code);
 }
 
+// TODO(sbc): There are no more uses to ATMAINS or addAtMain in emscripten.
+// We should look into removing these.
 global.ATMAINS = [];
 
 function addAtMain(code) {
+  assert(HAS_MAIN, 'addAtMain called but program has no main function');
   ATMAINS.push(code);
 }
 
@@ -1093,13 +912,6 @@ function makeRetainedCompilerSettings() {
 // In wasm, the heap size must be a multiple of 64KiB.
 const WASM_PAGE_SIZE = 65536;
 
-// Page size reported by some POSIX calls, mostly filesystem. This does not
-// depend on the memory page size which differs between wasm and asm.js, and
-// makes us report a consistent value despite the compile target. However,
-// perhaps we should unify all the page sizes (especially after fastcomp is
-// gone TODO).
-const POSIX_PAGE_SIZE = 16384;
-
 // Receives a function as text, and a function that constructs a modified
 // function, to which we pass the parsed-out name, arguments, and body of the
 // function. Returns the output of that function.
@@ -1144,7 +956,8 @@ function expectToReceiveOnModule(name) {
 function makeRemovedModuleAPIAssert(moduleName, localName) {
   if (!ASSERTIONS) return '';
   if (!localName) localName = moduleName;
-  return `if (!Object.getOwnPropertyDescriptor(Module, '${moduleName}')) {
+  return `
+if (!Object.getOwnPropertyDescriptor(Module, '${moduleName}')) {
   Object.defineProperty(Module, '${moduleName}', {
     configurable: true,
     get: function() {
@@ -1163,7 +976,7 @@ function makeModuleReceive(localName, moduleName) {
   if (expectToReceiveOnModule(moduleName)) {
     // Usually the local we use is the same as the Module property name,
     // but sometimes they must differ.
-    ret = "if (Module['" + moduleName + "']) " + localName + " = Module['" + moduleName + "'];";
+    ret = `\nif (Module['${moduleName}']) ${localName} = Module['${moduleName}'];`;
   }
   ret += makeRemovedModuleAPIAssert(moduleName, localName);
   return ret;
@@ -1192,12 +1005,20 @@ function makeModuleReceiveWithVar(localName, moduleName, defaultValue, noAssert)
 }
 
 function makeRemovedFSAssert(fsName) {
-  if (!ASSERTIONS) return;
+  assert(ASSERTIONS);
   const lower = fsName.toLowerCase();
-  if (SYSTEM_JS_LIBRARIES.indexOf('library_' + lower + '.js') >= 0) return '';
+  if (JS_LIBRARIES.includes('library_' + lower + '.js')) return '';
   return `var ${fsName} = '${fsName} is no longer included by default; build with -l${lower}.js';`;
 }
 
+function makeRemovedRuntimeFunction(name) {
+  assert(ASSERTIONS);
+  if (libraryFunctions.includes(name)) {
+    return '';
+  }
+  return `function ${name}() { abort('\`${name}\` is now a library function and not included by default; add it to your library.js __deps or to DEFAULT_LIBRARY_FUNCS_TO_INCLUDE on the command line'); }`;
+}
+
 // Given an array of elements [elem1,elem2,elem3], returns a string "['elem1','elem2','elem3']"
 function buildStringArray(array) {
   if (array.length > 0) {
@@ -1219,11 +1040,20 @@ function makeAsmImportsAccessInPthread(variable) {
     return `imports['${variable}']`;
   }
   // In non-MODULARIZE builds, can access the imports from global scope.
-  return variable;
+  return `self.${variable}`;
+}
+
+function _asmjsDemangle(symbol) {
+  if (symbol in WASM_SYSTEM_EXPORTS) {
+    return symbol;
+  }
+  // Strip leading "_"
+  assert(symbol.startsWith('_'));
+  return symbol.substr(1);
 }
 
 function hasExportedFunction(func) {
-  return Object.keys(EXPORTED_FUNCTIONS).indexOf(func) != -1;
+  return Object.keys(WASM_EXPORTS).includes(_asmjsDemangle(func));
 }
 
 // JS API I64 param handling: if we have BigInt support, the ABI is simple,
@@ -1254,6 +1084,26 @@ function sendI64Argument(low, high) {
   return low + ', ' + high;
 }
 
+// Any function called from wasm64 may have bigint args, this function takes
+// a list of variable names to convert to number.
+function from64(x) {
+  if (!MEMORY64) {
+    return '';
+  }
+  if (Array.isArray(x)) {
+    let ret = '';
+    for (e of x) ret += from64(e);
+    return ret;
+  } else {
+    return `${x} = Number(${x});`;
+  }
+}
+
+function to64(x) {
+  if (!MEMORY64) return x;
+  return `BigInt(${x})`;
+}
+
 // Add assertions to catch common errors when using the Promise object we
 // create on Module.ready() and return from MODULARIZE Module() invocations.
 function addReadyPromiseAssertions(promise) {
@@ -1278,7 +1128,7 @@ function addReadyPromiseAssertions(promise) {
 }
 
 function makeMalloc(source, param) {
-  if ('_malloc' in IMPLEMENTED_FUNCTIONS) {
+  if (hasExportedFunction('_malloc')) {
     return `_malloc(${param})`;
   }
   // It should be impossible to call some functions without malloc being
@@ -1291,3 +1141,21 @@ function makeMalloc(source, param) {
   }
   return `abort('malloc was not included, but is needed in ${source}. Adding "_malloc" to EXPORTED_FUNCTIONS should fix that. This may be a bug in the compiler, please file an issue.');`;
 }
+
+// Adds a call to runtimeKeepalivePush, if needed by the current build
+// configuration.
+// We skip this completely in MINIMAL_RUNTIME and also in builds that
+// don't ever need to exit the runtime.
+function runtimeKeepalivePush() {
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  return 'runtimeKeepalivePush();';
+}
+
+// Adds a call to runtimeKeepalivePush, if needed by the current build
+// configuration.
+// We skip this completely in MINIMAL_RUNTIME and also in builds that
+// don't ever need to exit the runtime.
+function runtimeKeepalivePop() {
+  if (MINIMAL_RUNTIME || (EXIT_RUNTIME == 0 && USE_PTHREADS == 0)) return '';
+  return 'runtimeKeepalivePop();';
+}
