@@ -11,12 +11,33 @@
 // and can be added/removed/renamed without fear of breaking out users.
 //
 
-// An array of all symbols exported from asm.js/wasm module.
-var MODULE_EXPORTS = [];
+// List of symbols exported from compiled code
+// These are raw symbol names and are not mangled to include the leading
+// underscore.
+var WASM_EXPORTS = [];
+
+// Similar to above but only includes the functions symbols.
+var WASM_FUNCTION_EXPORTS = [];
+
+// An array of all symbols exported from all the side modules specified on the
+// command line.
+// These are raw symbol names and are not mangled to include the leading
+// underscore.
+var SIDE_MODULE_EXPORTS = [];
+
+// All symbols imported by side modules.  These are symbols that the main
+// module (or other side modules) will need to provide.
+var SIDE_MODULE_IMPORTS = [];
+
+// Like EXPORTED_FUNCTIONS, but will not error if symbol is missing
+var EXPORT_IF_DEFINED = [];
 
 // stores the base name of the output file (-o TARGET_BASENAME.js)
 var TARGET_BASENAME = '';
 
+// stores the base name (with extension) of the output JS file
+var TARGET_JS_NAME = '';
+
 // Indicates that the syscalls (which we see statically) indicate that they need
 // full filesystem support. Otherwise, when just a small subset are used, we can
 // get away without including the full filesystem - in particular, if open() is
@@ -62,45 +83,39 @@ var EMBIND = 0;
 // Whether the main() function reads the argc/argv parameters.
 var MAIN_READS_PARAMS = 1;
 
-// List of functions implemented in compiled code; received from the backend.
-var IMPLEMENTED_FUNCTIONS = [];
-
 // Name of the file containing the Fetch *.fetch.js, if relevant
 var FETCH_WORKER_FILE = '';
 
 var WASI_MODULE_NAME = "wasi_snapshot_preview1";
 
-// Specifies a list of Emscripten-provided JS libraries to link against.
-// (internal, use -lfoo or -lfoo.js to link to Emscripten system JS libraries)
-var SYSTEM_JS_LIBRARIES = [];
+// List of JS libraries explictly linked against.  This includes JS system
+// libraries (specified via -lfoo or -lfoo.js) in addition to user libraries
+// passed via `--js-library`.  It does not include implicitly linked libraries
+// added by the JS compiler.
+var JS_LIBRARIES = [];
 
-// This will contain the emscripten version. You should not modify this. This
-// and the following few settings are useful in combination with
-// RETAIN_COMPILER_SETTINGS
+// This will contain the emscripten version. This can be useful in combination
+// with RETAIN_COMPILER_SETTINGS
 var EMSCRIPTEN_VERSION = '';
 
 // Will be set to 0 if -fno-rtti is used on the command line.
 var USE_RTTI = 1;
 
-// This will contain the optimization level (-Ox). You should not modify this.
+// This will contain the optimization level (-Ox).
 var OPT_LEVEL = 0;
 
-// This will contain the debug level (-gx). You should not modify this.
+// This will contain the debug level (-gx).
 var DEBUG_LEVEL = 0;
 
 // This will contain the shrink level (1 or 2 for -Os or -Oz, or just 0).
-// You should not modify this.
 var SHRINK_LEVEL = 0;
 
-// Whether we are profiling functions. You should not modify this.
-var PROFILING_FUNCS = 0;
+// Whether or not to emit the name section in the final wasm binaryen.
+var EMIT_NAME_SECTION = 0;
 
-// Whether we are emitting a symbol map. You should not modify this.
+// Whether we are emitting a symbol map.
 var EMIT_SYMBOL_MAP = 0;
 
-// tracks the list of EM_ASM signatures that are proxied between threads.
-var PROXIED_FUNCTION_SIGNATURES = [];
-
 // List of function explicitly exported by user on the command line.
 var USER_EXPORTED_FUNCTIONS = [];
 
@@ -119,9 +134,6 @@ var MEM_INIT_IN_WASM = 0;
 // This is set internally when needed (SINGLE_FILE)
 var SUPPORT_BASE64_EMBEDDING = 0;
 
-// the total initial wasm table size, only used in RELOCATABLE mode
-var WASM_TABLE_SIZE = 0;
-
 // the possible environments the code may run in.
 var ENVIRONMENT_MAY_BE_WEB = 1;
 var ENVIRONMENT_MAY_BE_WORKER = 1;
@@ -130,6 +142,7 @@ var ENVIRONMENT_MAY_BE_SHELL = 1;
 var ENVIRONMENT_MAY_BE_WEBVIEW = 1;
 
 // Whether to minify import and export names in the minify_wasm_js stage.
+// Currently always off for MEMORY64.
 var MINIFY_WASM_IMPORTS_AND_EXPORTS = 0;
 
 // Whether to minify imported module names.
@@ -144,7 +157,7 @@ var TARGET_NOT_SUPPORTED = 0x7FFFFFFF;
 // Wasm backend symbols that are considered system symbols and don't
 // have the normal C symbol name mangled applied (== prefix with an underscore)
 // (Also implicily on this list is any function that starts with string "dynCall_")
-var WASM_SYSTEM_EXPORTS = ['setTempRet0', 'getTempRet0', 'stackAlloc', 'stackSave', 'stackRestore'];
+var WASM_SYSTEM_EXPORTS = ['stackAlloc', 'stackSave', 'stackRestore'];
 
 // Internal: value of -flto argument (either full or thin)
 var LTO = 0;
@@ -178,12 +191,27 @@ var EXPECT_MAIN = 1;
 // MODULARIZE, and returned from the factory function.
 var EXPORT_READY_PROMISE = 1;
 
-var USE_LEGACY_DYNCALLS = 0;
-
 // struct_info that is either generated or cached
 var STRUCT_INFO = '';
 
-// If true, building against Emscripten's asm.js/wasm heap memory profiler.
+// If true, building against Emscripten's wasm heap memory profiler.
 var MEMORYPROFILER = 0;
 
 var GENERATE_SOURCE_MAP = 0;
+
+// Memory layout.  These are only used/set in RELOCATABLE builds.  Otherwise
+// memory layout is fixed in the wasm binary at link time.
+var STACK_BASE = 0;
+var STACK_MAX = 0;
+var HEAP_BASE = 0;
+
+// Used internally. set when there is a main() function.
+// Also set when in a linkable module, as the main() function might
+// arrive from a dynamically-linked library, and not necessarily
+// the current compilation unit.
+// Also set for STANDALONE_WASM since the _start function is needed to call
+// static ctors, even if there is no user main.
+var HAS_MAIN = 0;
+
+// Set to true if we are linking as C++ and including C++ stdlibs
+var LINK_AS_CXX = 0;
