@@ -8,6 +8,7 @@
 #include <emscripten/fetch.h>
 #include <emscripten/html5.h>
 #include <emscripten/threading.h>
+#include <emscripten/console.h>
 #include <math.h>
 #include <memory.h>
 #include <stdlib.h>
@@ -18,43 +19,40 @@ extern "C" {
 // Uncomment the following and clear the cache with emcc --clear-cache to rebuild this file to
 // enable internal debugging. #define FETCH_DEBUG
 
-struct __emscripten_fetch_queue {
+static void fetch_free(emscripten_fetch_t* fetch);
+
+// APIs defined in JS
+void emscripten_start_fetch(emscripten_fetch_t* fetch);
+int32_t _emscripten_fetch_get_response_headers_length(int32_t fetchID);
+int32_t _emscripten_fetch_get_response_headers(int32_t fetchID, int32_t dst, int32_t dstSizeBytes);
+void _emscripten_fetch_free(unsigned int);
+
+struct emscripten_fetch_queue {
   emscripten_fetch_t** queuedOperations;
   int numQueuedItems;
   int queueSize;
 };
 
-static void fetch_free(emscripten_fetch_t* fetch);
+emscripten_fetch_queue* _emscripten_get_fetch_queue() {
+  static thread_local emscripten_fetch_queue g_queue;
 
-extern "C" {
-void emscripten_start_fetch(emscripten_fetch_t* fetch);
-__emscripten_fetch_queue* _emscripten_get_fetch_work_queue();
-
-__emscripten_fetch_queue* _emscripten_get_fetch_queue() {
-  __emscripten_fetch_queue* queue = _emscripten_get_fetch_work_queue();
-  if (!queue->queuedOperations) {
-    queue->queueSize = 64;
-    queue->numQueuedItems = 0;
-    queue->queuedOperations =
-      (emscripten_fetch_t**)malloc(sizeof(emscripten_fetch_t*) * queue->queueSize);
+  if (!g_queue.queuedOperations) {
+    g_queue.queueSize = 64;
+    g_queue.numQueuedItems = 0;
+    g_queue.queuedOperations =
+      (emscripten_fetch_t**)malloc(sizeof(emscripten_fetch_t*) * g_queue.queueSize);
   }
-  return queue;
-}
-
-int32_t _emscripten_fetch_get_response_headers_length(int32_t fetchID);
-int32_t _emscripten_fetch_get_response_headers(int32_t fetchID, int32_t dst, int32_t dstSizeBytes);
-void _emscripten_fetch_free(unsigned int);
+  return &g_queue;
 }
 
 void emscripten_proxy_fetch(emscripten_fetch_t* fetch) {
   // TODO: mutex lock
-  __emscripten_fetch_queue* queue = _emscripten_get_fetch_queue();
+  emscripten_fetch_queue* queue = _emscripten_get_fetch_queue();
   //	TODO handle case when queue->numQueuedItems >= queue->queueSize
   queue->queuedOperations[queue->numQueuedItems++] = fetch;
 #ifdef FETCH_DEBUG
-  EM_ASM(console.log('Queued fetch to fetch-worker to process. There are now ' + $0 +
-                     ' operations in the queue.'),
-    queue->numQueuedItems);
+  emscripten_console_logf("Queued fetch to fetch-worker to process. There are "
+                          "now %d operations in the queue.", queue->numQueuedItems);
 #endif
   // TODO: mutex unlock
 }
@@ -80,11 +78,7 @@ emscripten_fetch_t* emscripten_fetch(emscripten_fetch_attr_t* fetch_attr, const
   const bool isMainBrowserThread = emscripten_is_main_browser_thread() != 0;
   if (isMainBrowserThread && synchronous && (performXhr || readFromIndexedDB || writeToIndexedDB)) {
 #ifdef FETCH_DEBUG
-    EM_ASM(
-      err(
-        'emscripten_fetch("' + UTF8ToString($0) +
-        '") failed! Synchronous blocking XHRs and IndexedDB operations are not supported on the main browser thread. Try dropping the EMSCRIPTEN_FETCH_SYNCHRONOUS flag, or run with the linker flag --proxy-to-worker to decouple main C runtime thread from the main browser thread.'),
-      url);
+    emscripten_console_errorf("emscripten_fetch('%s') failed! Synchronous blocking XHRs and IndexedDB operations are not supported on the main browser thread. Try dropping the EMSCRIPTEN_FETCH_SYNCHRONOUS flag, or run with the linker flag --proxy-to-worker to decouple main C runtime thread from the main browser thread.", url);
 #endif
     return 0;
   }
@@ -165,7 +159,7 @@ emscripten_fetch_t* emscripten_fetch(emscripten_fetch_attr_t* fetch_attr, const
       || (synchronous &&
            (readFromIndexedDB || writeToIndexedDB))) // Synchronous IndexedDB access needs proxying
   {
-    emscripten_atomic_store_u32(&fetch->__proxyState, 1); // sent to proxy worker.
+    fetch->__proxyState = 1; // sent to proxy worker.
     emscripten_proxy_fetch(fetch);
 
     if (synchronous)
@@ -180,13 +174,13 @@ EMSCRIPTEN_RESULT emscripten_fetch_wait(emscripten_fetch_t* fetch, double timeou
 #if __EMSCRIPTEN_PTHREADS__
   if (!fetch)
     return EMSCRIPTEN_RESULT_INVALID_PARAM;
-  uint32_t proxyState = emscripten_atomic_load_u32(&fetch->__proxyState);
+  uint32_t proxyState = fetch->__proxyState;
   if (proxyState == 2)
     return EMSCRIPTEN_RESULT_SUCCESS; // already finished.
   if (proxyState != 1)
     return EMSCRIPTEN_RESULT_INVALID_PARAM; // the fetch should be ongoing?
 #ifdef FETCH_DEBUG
-  EM_ASM(console.log('fetch: emscripten_fetch_wait..'));
+  emscripten_console_log("fetch: emscripten_fetch_wait..");
 #endif
   if (timeoutMsecs <= 0)
     return EMSCRIPTEN_RESULT_TIMED_OUT;
@@ -195,15 +189,14 @@ EMSCRIPTEN_RESULT emscripten_fetch_wait(emscripten_fetch_t* fetch, double timeou
       int ret = emscripten_futex_wait(&fetch->__proxyState, proxyState, timeoutMsecs);
       if (ret == -ETIMEDOUT)
         return EMSCRIPTEN_RESULT_TIMED_OUT;
-      proxyState = emscripten_atomic_load_u32(&fetch->__proxyState);
+      proxyState = fetch->__proxyState;
     } else {
-      EM_ASM({console.error(
-        'fetch: emscripten_fetch_wait failed: main thread cannot block to wait for long periods of time! Migrate the application to run in a worker to perform synchronous file IO, or switch to using asynchronous IO.')});
+      emscripten_console_error("fetch: emscripten_fetch_wait failed: main thread cannot block to wait for long periods of time! Migrate the application to run in a worker to perform synchronous file IO, or switch to using asynchronous IO.");
       return EMSCRIPTEN_RESULT_FAILED;
     }
   }
 #ifdef FETCH_DEBUG
-  EM_ASM(console.log('fetch: emscripten_fetch_wait done..'));
+  emscripten_console_log("fetch: emscripten_fetch_wait done..");
 #endif
 
   if (proxyState == 2)
@@ -217,8 +210,7 @@ EMSCRIPTEN_RESULT emscripten_fetch_wait(emscripten_fetch_t* fetch, double timeou
     return EMSCRIPTEN_RESULT_TIMED_OUT /*Main thread testing completion with sleep=0msecs*/;
   else {
 #ifdef FETCH_DEBUG
-    EM_ASM(console.error(
-      'fetch: emscripten_fetch_wait() cannot stop to wait when building without pthreads!'));
+    emscripten_console_error("fetch: emscripten_fetch_wait() cannot stop to wait when building without pthreads!");
 #endif
     return EMSCRIPTEN_RESULT_FAILED /*Main thread cannot block to wait*/;
   }
@@ -230,7 +222,7 @@ EMSCRIPTEN_RESULT emscripten_fetch_close(emscripten_fetch_t* fetch) {
     return EMSCRIPTEN_RESULT_SUCCESS; // Closing null pointer is ok, same as with free().
 
 #if __EMSCRIPTEN_PTHREADS__
-  emscripten_atomic_store_u32(&fetch->__proxyState, 0);
+  fetch->__proxyState = 0;
 #endif
   // This function frees the fetch pointer so that it is invalid to access it anymore.
   // Use a few key fields as an integrity check that we are being passed a good pointer to a valid
